unit KelHexEditor;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Grids;

const
  BPL = 16;
  OffsetFore = clBlack;
  OffsetBack = clWhite;
  HexEvenFore = clBlack;
  HexEvenBack = clWhite;
  HexOddFore  = clBlue;
  HexOddBack  = clWhite;

  SelectedBack = clBlue;
  SelectedFore = clWhite;

  DataBack = clWhite;
  DataFore = clBlack;

  HexStart = 2;
  HexEnd   = 49;
  DataStart= 50;
  DataEnd  = 66;
  WM_STATECHANGED = WM_USER +2 ;


type
  TCaretType = (ctNone, ctHex, ctData);

  TKelHexEditor = class(TCustomGrid)
  private
    { Private declarations }
    FCalcSizes : Boolean;

    FOffsetFont : TFont;
    FOffsetHeight : Integer;
    FOffsetWidth : Integer;

    FHexFont : TFont;
    FHexHeight : Integer;
    FHexWidth : Integer;

    FDataFont : TFont;
    FDataHeight : Integer;
    FDataWidth : Integer;

    FMaxWidth : Integer;
    FMaxHeight : Integer;

    FIsSelecting : Boolean;
    FSelST , FSelPO , FSelEN : Integer;

    DataBuffer : Pointer;
    DataSize : Integer;
    DataLines : Integer;

    CaretType : TCaretType;
    Caret : TBitmap;

    function GetLines(FromSize : Integer) : Integer;
    function GetOtherCol(ACol : Integer) : Integer;
    function GetPosAtCursor(ACol, ARow : Integer) : Integer;
    function GetColAtCursor(ACol : Integer) : Integer;
    procedure CalcSizes;
    procedure CreateHexCaret;
    procedure CreateDataCaret;
    procedure RedrawPos ( aFrom , aTo : Integer );
  protected
    { Protected declarations }
    procedure Select ( aCurCol , aCurRow , aNewCol , aNewRow : Integer );
    procedure CheckCell(var ACol, ARow : Integer);
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    procedure SetOffsetFont(NewValue : TFont);
    procedure SetHexFont(NewValue : TFont);
    procedure SetDataFont(NewValue : TFont);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    function  SelectCell(ACol, ARow : Integer) : Boolean; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure StateNotification;
    procedure WMHScroll(var Msg: TWMVScroll); message WM_HSCROLL;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    function IsSelected ( aPos : Integer ) : Boolean;
  public
    { Public declarations }

    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;

    procedure GoNextColumn;
    procedure GoNextLine;
    procedure GoPrevColumn;
    procedure GoPrevLine;
    procedure GoHome;
    procedure GoEnd;
    procedure GoPageDown;
    procedure GoPageUp;
    procedure ResetSelection ( aDraw : Boolean);
  published
    { Published declarations }
    property Align;
    property Anchors;
    property BiDiMode;
    property BorderStyle;
    property Color;
    property ColCount;
    property Constraints;
    property Ctl3D;
    property DefaultColWidth;
    property DefaultRowHeight;
    property DefaultDrawing;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FixedColor;
    property FixedCols;
    property RowCount;
    property FixedRows;
    property Font;
    property GridLineWidth;
    property Options;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property VisibleColCount;
    property VisibleRowCount;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnStartDock;
    property OnStartDrag;


    property FontOffset : TFont read FOffsetFont write SetOffsetFont;
    property FontHex : TFont read FHexFont write SetHexFont;
    property FontData : TFont read FDataFont write SetDataFont;
  end;

procedure Register;

implementation

procedure Register;
  begin
    RegisterComponents('Samples', [TKelHexEditor]);
  end;

function Max(const I : array of Integer) : Integer;
  var
    N : Integer;
  begin
    Result := -1;
    for N:=0 to (sizeof(I) div sizeof(Integer))-1 do
      if I[N] > Result then
        Result := I[N];
  end;

function Min(const I : array of Integer) : Integer;
  var
    N : Integer;
  begin
    Result := 32000;
    for N:=0 to (sizeof(I) div sizeof(Integer))-1 do
      if I[N] < Result then
        Result := I[N];
  end;


constructor TKelHexEditor.Create(AOwner : TComponent);
  var
    P : PChar;
    n : Integer;
  begin
    inherited;

    DefaultDrawing := False;
    Options := [goVertLine, goHorzLine,goTabs,gOThumbTracking];    

    DataSize := 1035;
    DataLines := GetLines(DataSize);

    GetMem(DataBuffer, DataSize);

    P := DataBuffer;
    for n:=1 to DataSize do
    begin
      P^ := Chr(n and $FF);
      Inc(P);
    end;

    FCalcSizes := True;

    FOffsetWidth := 0;
    FOffsetHeight := 0;
    FOffsetFont := TFont.Create;

    FHexWidth := 0;
    FHexHeight := 0;
    FHexFont := TFont.Create;

    FDataWidth := 0;
    FDataHeight := 0;
    FDataFont := TFont.Create;

    FIsSelecting := False;


    Caret := TBitmap.Create;
    CaretType := ctNone;

    SetOffsetFont(nil);
    SetHexFont(nil);
    SetDataFont(nil);

  end;

destructor TKelHexEditor.Destroy;
  begin
    inherited;

    FreeMem(DataBuffer, DataSize);

    FOffsetFont.Free;
    FHexFont.Free;
    FDataFont.Free;

    Caret.Free;
  end;

function  TKelHexEditor.GetLines(FromSize : Integer) : Integer;
  begin
    Result := FromSize div BPL;

    if FromSize mod BPL > 0 then
      Inc(Result);
  end;

procedure TKelHexEditor.CalcSizes;
  var
    N : Integer;
  begin
//    SetOffsetFont(nil);
//    SetHexFont(nil);
//    SetDataFont(nil);

    RowCount := 1 + GetLines(DataSize);
    ColCount := 67;
    FixedRows := 1;
    FixedCols := 2;

    if FHexWidth>0 then
      for N:=HexStart to HexEnd do
      begin
        ColWidths[N] := FHexWidth;
      end;

    if FDataWidth>0 then
      for N:=DataStart to DataEnd do
      begin
        ColWidths[N] := FDataWidth;
      end;
  end;



const
  HexStr = '0123456789ABCDEF -';

procedure TKelHexEditor.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
  var
    Text : String;
    TextColor : TColor;
    BackColor : TColor;
    I : Integer;
    H : Integer;
    P : PChar;
    B : Byte;
    C : Integer;

  procedure _PrintText;
    begin
      Windows.SetTextColor(Canvas.Handle, TextColor);
      Windows.SetBkColor(Canvas.Handle, BackColor);
      Windows.SetTextAlign(Canvas.Handle, TA_CENTER or VTA_CENTER);
      Windows.ExtTextOut(
          Canvas.Handle,
          (ARect.Left + ARect.Right) shr 1,
          (ARect.Top + ARect.Bottom) shr 1 - H shr 1,
          ETO_CLIPPED or ETO_OPAQUE,
          @ARect,
          PChar(Text),
          Length(Text),
          nil);
    end;

    begin
      if FCalcSizes then
      begin
        FCalcSizes := False;
        CalcSizes;
      end;

      Text := ' ';
      case ACol of
        0 :
          begin
            if ARow>0 then
            begin
              I := BPL * (ARow-1);
              Text := Format('%8xh', [I]);
              TextColor := OffsetFore;
              BackColor := OffsetBack;
              Canvas.Font.Assign(FontOffset);
              H := FOffsetHeight;
              _PrintText;
            end;
          end;

        1 :
          begin
            if ARow>0 then
            begin
              I := BPL * (ARow-1);
              Text := Format('%8dd', [I]);
              TextColor := OffsetFore;
              BackColor := OffsetBack;
              Canvas.Font.Assign(FontOffset);
              H := FOffsetHeight;
              _PrintText;
            end;
          end;

        HexStart..HexEnd :
          begin
            if ARow = 0 then
            begin
              C := (ACol-2) mod 3;

              if ( (ACol-2) div 3) mod 2 = 0 then
              begin
                TextColor := HexEvenFore;
                BackColor := HexEvenBack;
              end else
              begin
                TextColor := HexOddFore;
                BackColor := HexOddBack;
              end;

              Text := IntToStr( (ACol-2) div 3);
              if C = 2 then
              begin
                H := FOffsetHeight-1;
                Windows.SetTextColor(Canvas.Handle, TextColor);
                Windows.SetBkColor(Canvas.Handle, BackColor);
                Windows.SetTextAlign(Canvas.Handle, TA_CENTER or VTA_CENTER);
                ARect.Left := ARect.Left - FHexWidth*2 + 1;
                ARect.Right := ARect.Right - 1;
                //ARect.Right := ARect.Left + FHexWidth*3;
                Windows.ExtTextOut(
                    Canvas.Handle,
                    (ARect.Left + ARect.Right) shr 1,
                    (ARect.Top + ARect.Bottom) shr 1 - H shr 1,
                    ETO_CLIPPED or ETO_OPAQUE,
                    @ARect,
                    PChar(Text),
                    Length(Text),
                    nil);
              end;
            end else
            begin
              I := GetPosAtCursor(ACol, ARow);
              if I<DataSize then
              begin
                P := DataBuffer;
                P := P + I;

                B := Byte(P^);
                C := (ACol-2) mod 3;

                if ( (ACol-2) div 3) mod 2 = 0 then
                begin
                  TextColor := HexEvenFore;
                  BackColor := HexEvenBack;
                end else
                begin
                  TextColor := HexOddFore;
                  BackColor := HexOddBack;
                end;

                if ACol = (DataStart-1) then BackColor := OffsetBack;
                case C of
                  0 :
                    begin
                      B := (B and $F0) shr 4;
                    end;

                  1 :
                    begin
                      B := (B and $0F);
                    end;

                  2 :
                    begin
                      if ACol=25 then
                        B := 17
                      else
                        B := 16;
                      BackColor := OffsetBack;
                      TextColor := OffsetFore;
                    end;
                end;

                if (IsSelected( GetPosAtCursor(ACol, ARow) ) ) and (C<>2) then
                begin
                  BackColor := SelectedBack;
                  TextColor := SelectedFore;
                end;

                Text := HexStr[B+1];
                Canvas.Font.Assign(FontHex);
                H := FHexHeight;
                _PrintText;
              end;
            end;
          end;

        DataStart..DataEnd :
          begin
            if ARow=0 then
            begin
              I := ACol - DataStart;
              if I>8 then
              begin
                Dec(I);
                Text := Format('%d', [I]);
              end else if I<8 then 
                Text := Format('%d', [I])
              else
                Text := ' ';
              BackColor := OffsetBack;
              TextColor := OffsetFore;
              Canvas.Font.Assign(FontOffset);
              H := FDataHeight;
              _PrintText;
            end else
            if ARow>0 then
            begin
              I := GetPosAtCursor(ACol, ARow);
              if I<DataSize then
              begin
                P := DataBuffer;
                P := P + I;

                if ACol=DataStart+8 then
                begin
                  B := Ord('-');
                  BackColor := OffsetBack;
                  TextColor := OffsetFore;
                end
                else
                  B := Byte(P^);

                BackColor := DataBack;
                TextColor := DataFore;

                if IsSelected( GetPosAtCursor(ACol, ARow) ) then
                begin
                  BackColor := SelectedBack;
                  TextColor := SelectedFore;
                end;

                Text := Chr(B);
                Canvas.Font.Assign(FontData);
                H := FDataHeight;
                _PrintText;
              end;
            end;
          end;
      end;

      // Do the Cursor Thing
      if ARow=Row then
      begin
        if (ACol = Col) then
        begin
          if ACol>=DataStart then
          begin
            CreateDataCaret;
            Windows.SetCaretPos(ARect.Left, ARect.Top);
          end else
          begin
            CreateHexCaret;
            Windows.SetCaretPos(ARect.Left, ARect.Top);
          end;
        end;

        if (ACol = GetOtherCol(Col)) then
        begin
          if ACol >= DataStart then
            Canvas.DrawFocusRect(Rect(aRect.Left,aRect.Top,aRect.Left+FDataWidth,aRect.Bottom))
          else
            Canvas.DrawFocusRect(Rect(aRect.Left,aRect.Top,aRect.Right,aRect.Bottom))
        end;
      end;
    end;

procedure TKelHexEditor.SetOffsetFont(NewValue : TFont);
  var
    Bitmap : TBitmap;
  begin
    if NewValue<>nil then
    begin
      FOffsetFont.Assign(NewValue);
    end;

    Bitmap := TBitmap.Create;
    Bitmap.Canvas.Font.Assign(FOffsetFont);
    FOffsetWidth := Bitmap.Canvas.TextWidth('88888888W ')+2;
    FOffsetHeight:= Bitmap.Canvas.TextHeight('88888888W ')+2;
    Bitmap.Free;

    FMaxWidth := Max([FOffsetWidth, FHexWidth, FDataWidth]);
    FMaxHeight:= Max([FOffsetHeight, FHexHeight, FDataHeight]);

    CalcSizes;

    if NewValue<>nil then Refresh;
  end;


procedure TKelHexEditor.SetHexFont(NewValue : TFont);
  var
    Bitmap : TBitmap;
  begin
    if NewValue<>nil then
    begin
      FHexFont.Assign(NewValue);
    end;

    Bitmap := TBitmap.Create;
    Bitmap.Canvas.Font.Assign(FHexFont);
    FHexWidth := Bitmap.Canvas.TextWidth('F')+2;
    FHexHeight:= Bitmap.Canvas.TextHeight('F')+2;
    Bitmap.Free;

    FMaxWidth := Max([FOffsetWidth, FHexWidth, FDataWidth]);
    FMaxHeight:= Max([FOffsetHeight, FHexHeight, FDataHeight]);

    CalcSizes;

    if NewValue<>nil then Refresh;
  end;


function TKelHexEditor.GetOtherCol(ACol : LongInt) : Integer;
  begin
    if ACol>=DataStart then
    begin
      // Convert from Data Area
      Result := ACol-DataStart;
      if Result>=8 then Dec(Result);

      Result := (Result*3)+2;
    end else
    begin
      // Convert from Hex Area
      Result := (ACol-1) div 3;
      if Result>=8 then Inc(Result);
      Result := Result + DataStart;
    end;
  end;


procedure TKelHexEditor.SetDataFont(NewValue : TFont);
  var
    Bitmap : TBitmap;
  begin
    if NewValue<>nil then
    begin
      FDataFont.Assign(NewValue);
    end;

    Bitmap := TBitmap.Create;
    Bitmap.Canvas.Font.Assign(FDataFont);
    FDataWidth := Bitmap.Canvas.TextWidth('W')+4;
    FDataHeight:= Bitmap.Canvas.TextHeight('W')+4;
    Bitmap.Free;

    FMaxWidth := Max([FOffsetWidth, FHexWidth, FDataWidth]);
    FMaxHeight:= Max([FOffsetHeight, FHexHeight, FDataHeight]);

    CalcSizes;

    if NewValue<>nil then Refresh;
  end;

procedure TKelHexEditor.CreateHexCaret;
  begin
    if CaretType<>ctHex then
    begin
      Windows.DestroyCaret;

      Canvas.Font.Assign(FontHex);
      Caret.Width :=   FHexWidth;
      Caret.Height :=  RowHeights[1];

      Windows.CreateCaret(
        Handle,
        Caret.Handle,
        Caret.Width,
        Caret.Height);

      CaretType := ctHex;
    end;
  end;

procedure TKelHexEditor.CreateDataCaret;
  begin
    if CaretType<>ctData then
    begin
      Windows.DestroyCaret;

      Canvas.Font.Assign(FontData);
      Caret.Width :=   FDataWidth;
      Caret.Height :=  RowHeights[1];

      Windows.CreateCaret(
        Handle,
        Caret.Handle,
        Caret.Width,
        Caret.Height);

      CaretType := ctData;
    end;
  end;

function TKelHexEditor.GetColAtCursor(ACol : Integer) : Integer;
  begin
    if ACol>=DataStart then
    begin
      Result := ACol-DataStart;
      if Result>=8 then Dec(Result);
    end else
    begin
      Result := (ACol-1) div 3;
    end;
  end;

function TKelHexEditor.GetPosAtCursor(ACol, ARow : Integer) : Integer;
  begin
    Result := (ARow-1) * BPL;
    Result := Result + GetColAtCursor(ACol);
  end;

procedure TKelHexEditor.GoNextLine;
  var
    I : Integer;
  begin
    I := Row;
    Inc(I);

    if SelectCell(Col, I) then
    begin
      Row := I;
    end;
  end;

function TKelHexEditor.SelectCell(ACol, ARow : Integer) : Boolean;
  begin
    Result := False;
    if
      (ACol>=HexStart) and (ACol<=DataEnd) and
      (ARow>=1) and (ARow<RowCount) then
    begin
      if GetPosAtCursor(ACol, ARow)<=DataSize then
        Result := True;
    end;

    if fIsSelecting
    then
        Select ( Col , Row , aCol , aRow )
    else
        ResetSelection( True);
  end;


procedure TKelHexEditor.GoPrevLine;
  var
    I : Integer;
  begin
    I := Row;
    Dec(I);

    if SelectCell(Col, I) then
    begin
      Row := I;
    end;
  end;

procedure TKelHexEditor.GoNextColumn;
  var
    I : Integer;
    C : Integer;
    R : Integer;
  begin
    I := Col + 1;

    if I<DataStart then
    begin
      C := (I-2) mod 3;
      if C = 2 then Inc(I);
      if I>=DataStart then
      begin
        Col := HexStart;
        GoNextLine;
        Exit;
      end else
      begin
        R := Row;
        Self.CheckCell(I, R);
      end;
    end else
    begin
      if I=58 then Inc(I);
      if I>DataEnd then
      begin
        Col := DataStart;
        GoNextLine;
        Exit;
      end;
    end;
    Col := I;
  end;


procedure TKelHexEditor.GoHome;
  begin
    if Col>=DataStart then
    begin
      Col := DataStart;
    end else
    begin
      Col := HexStart;
    end;
  end;

procedure TKelHexEditor.GoEnd;
  var
    I  : Integer;
    ARow : Integer;
    ACol : Integer;
  begin
    ARow := Row;
    ACol := Col;
    if ACol>=DataStart then
      ACol := DataEnd
    else
      ACol := HexEnd-1;

    CheckCell(ACol, ARow);
    Col := ACol;
    Row := ARow;
  end;


procedure TKelHexEditor.CheckCell(var ACol, ARow : Integer);
  var
    I : Integer;
    C : Integer;
  begin
    if ARow>=RowCount then ARow := RowCount-1;
    if ARow<1 then ARow := 1;

    if ARow = RowCount-1 then
    begin
      I := DataSize mod BPL;
      if ACol>DataStart then
      begin
        C := DataStart+I;
        if ACol>C then
          ACol := C;
      end else
      begin
        C := HexStart+I*3-2;
        if ACol>C then
          ACol := C;
      end;
    end;
  end;


procedure TKelHexEditor.GoPageDown;
  var
    LRow : Integer;
    LCol : Integer;
  begin
    LRow := Row + Self.VisibleRowCount;
    LCol := Col;

    CheckCell(LCol, LRow);
    Col := LCol;
    Row := LRow;
  end;


procedure TKelHexEditor.GoPageUp;
  var
    LRow : Integer;
    LCol : Integer;
  begin
    LRow := Row - Self.VisibleRowCount;
    LCol := Col;

    CheckCell(LCol, LRow);
    Col := LCol;
    Row := LRow;
  end;

procedure TKelHexEditor.GoPrevColumn;
  var
    I : Integer;
    C : Integer;
  begin
    I := Col-1;

    if I<HexEnd then
    begin
      C := (I-HexStart) mod 3;
      if C = 2 then Dec(I);

      if I<HexStart then
      begin
        GoPrevLine;
        Col := HexEnd-1;
        Exit;
      end;
    end else
    begin
      if I=58 then Dec(I);
      if I<DataStart then
      begin
        Col := DataEnd;
        GoPrevLine;
        Exit;
      end;
    end;
    Col := I;
  end;

procedure TKelHexEditor.KeyDown(var Key: Word; Shift: TShiftState);
  var
    I : Integer;
  begin
    InvalidateCell(Col, Row);
    InvalidateCell(GetOtherCol(Col),Row);

    if ssShift in Shift then
    begin
      if not FIsSelecting then
      begin
        FIsSelecting := True;
      end;
    end else
      if FIsSelecting then
      begin
        ResetSelection(True);
      end;


    case Key of
      VK_HOME  :
        begin
          if ssCtrl in Shift then
            Row := 1;
          GoHome;
        end;

      VK_END   :
        begin
          if ssCtrl in Shift then
          begin
            GoHome;
            Row := RowCount-1;
          end;
          GoEnd;
        end;

      VK_RIGHT : GoNextColumn;
      VK_LEFT  : GoPrevColumn;
      VK_DOWN  : GoNextLine;
      VK_UP    : GoPrevLine;
      VK_NEXT  : GoPageDown;
      VK_PRIOR : GoPageUp;

      VK_TAB :
        begin
          I := GetOtherCol(Col);
          Col := I;
        end;
    end;

    if (ssShift in Shift) and (FIsSelecting) then
    begin
      FSelEn := GetPosAtCursor(Col, Row);
    end else
    begin
      InvalidateCell(Col, Row);
      InvalidateCell(GetOtherCol(Col),Row);
    end;
  end;

procedure TKelHexEditor.MouseMove(Shift: TShiftState; X, Y: Integer);
  var
    Coord : TGridCoord;
  begin
    inherited;
    if (ssLeft in Shift) then
    begin
      //TCustomGrid
      FIsSelecting := True;
      try
        Coord := MouseCoord( X, Y );
        CheckCell(Coord.X, Coord.Y);
        SelectCell(Coord.X, Coord.Y);
      except
        Application.MessageBox('Error','Error',IDOK);
      end;
    end
  end;

procedure TKelHexEditor.MouseDown(
        Button: TMouseButton;
        Shift: TShiftState;
        X, Y: Integer);
  var
    Key : Word;
  begin
    inherited;

    if ssLeft in Shift then
    begin
      InvalidateCell(Col, Row);
      InvalidateCell(GetOtherCol(Col),Row);



      GoPrevColumn;
      GoNextColumn;

      if Button = mbLeft then
      begin
        ResetSelection( True );
        //if not (ssDouble in Shift) then fIsSelecting := True;
      end;

      InvalidateCell(Col, Row);
      InvalidateCell(GetOtherCol(Col),Row);
    end;
  end;


procedure TKelHexEditor.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
  var
    Coord : TGridCoord;
    PosSource, PosDestin : Integer;
  begin
    if (Button = mbRight) and (FIsSelecting) then
    begin
      Coord := MouseCoord( X, Y );
      if IsSelected( GetPosAtCursor(Coord.X,Coord.Y) ) then
      begin
        Application.MessageBox('A','A',IDOK);
      end;
    end else
      inherited;

  end;




procedure TKelHexEditor.ResetSelection ( aDraw : Boolean );
  var
    pOldFrom , pOldTo : Integer;
  begin
    FIsSelecting := False;
    pOldFrom := fSelST;
    pOldTo := fSelEN;
    fSelST := -1;
    fSelPO := -1;
    fSelEN := -1;
    if aDraw then
      RedrawPos ( pOldFrom, pOldTo );
    StateNotification;
  end;


procedure TKelHexEditor.StateNotification;
  begin
    if HandleAllocated then
      PostMessage ( Handle , WM_STATECHANGED , 7 , 7 );
  end;

procedure TKelHexEditor.WMVScroll(var Msg: TWMVScroll);
  var
    pRC : TRect;
  begin
    inherited;
    pRC := CellRect ( Col , Row );
    if pRC.Left+pRC.Bottom = 0 then
      SetCaretPos ( -50 , -50)
    else
      SetCaretPos ( pRC.Left , pRC.Top);
  end;

procedure TKelHexEditor.WMHScroll(var Msg: TWMHScroll);
  var
    pRC : TRect;
  begin
    inherited;
    pRC := CellRect ( Col , Row );
    if pRC.Left+pRC.Bottom = 0 then
      SetCaretPos ( -50 , -50)
    else
      SetCaretPos ( pRC.Left , pRC.Top);
  end;



function TKelHexEditor.IsSelected ( aPos : Integer ) :Boolean;
  var
    A, B : Integer;
  begin
    A := Min([fSelST,fSelEn]);
    B := Max([fSelST,fSelEn]);
    Result := False;
    if (aPos >= A) and (aPos <= B) then
      Result := True;
  end;


procedure TKelHexEditor.RedrawPos ( aFrom , aTo : Integer );
  var
    pR : TRect;
    A, B : Integer;
    AC, BC : Integer;
  begin
    A := Min([aFrom, aTo]);
    B := Max([aFrom, aTo]);

    if A mod BPL <> 0 then AC := 1;
    if B mod BPL <> 0 then BC := 1;

    A := A div BPL - AC;
    B := B div BPL + BC;

    pR := BoxRect ( 2 , A , ColCount -1 , B );
    InvalidateRect ( Handle , @pR , False );
  end;

procedure TKelHexEditor.Select ( aCurCol , aCurRow , aNewCol , aNewRow : Integer );
  var
    pOST , pOEN , pNAP : Integer;
  begin
    pOEN := fSelEN;
    pOST := fSelST;
    pNAP := GetPosAtCursor ( aNewCol , aNewRow  );
    if fSelPO = -1 then
    begin
      fSelPO := GetPosAtCursor ( aCurCol , aCurRow );
      pOST := pNAP;
      pOEN := pNAP;
      fSelST := Min ( [pOST , fSelPO]);
      fSelEN := Max ( [fSelPO , pOEN] );
      RedrawPos ( fSelST , fSelEN );
    end else
    begin
        fSelEN := pNAP;
        fSelST := fSelPO;
        RedrawPos ( Min ( [fSelST , pOST] ) , Max ( [fSelST , pOST] ));
        RedrawPos ( Min ( [fSelEN , pOEN] ) , Max ( [fSelEN , pOEN] ));
      //end;
    end;
    StateNotification;
  end;



end.

